import struct
import time
from typing import List, Dict, Tuple

import serial
from serial.tools.list_ports import comports
import threading
import smbus2

from abc import ABC, abstractmethod

from curverunner.util import INT16_MAX, map_value, unsigned_to_signed2, unsigned_to_signed4


class CurverunnerComm(ABC):
    """
    Abstract base class for communication with CurveRunner device.

    Methods to read and write 1, 2, and 4 byte values to/from specified addresses.
    """

    @abstractmethod
    def write1(self, addr: int, val: int) -> None:
        pass

    @abstractmethod
    def write2(self, addr: int, val: int) -> None:
        pass

    @abstractmethod
    def write4(self, addr: int, val: int) -> None:
        pass

    @abstractmethod
    def read1(self, addr: int) -> int:
        pass

    @abstractmethod
    def read2(self, addr: int) -> int:
        pass

    @abstractmethod
    def read4(self, addr: int) -> int:
        pass


class CurverunnerCommI2C(CurverunnerComm):
    def __init__(self, bus: smbus2.SMBus, addr: int = 0x08):
        self._bus = bus
        self._addr = addr

    def _attempt_write(self, reg, data):
        for _ in range(5):
            try:
                self._bus.write_i2c_block_data(self._addr, reg, data)
                break
            except OSError:
                time.sleep(0.02)

    def _attempt_read(self, reg, size):
        byte_data = None
        for _ in range(5):
            try:
                byte_data = bytes(self._bus.read_i2c_block_data(self._addr, reg, size))
                break
            except OSError:
                time.sleep(0.02)
        else:
            raise RuntimeError("Failed to read from I2C device")
        return byte_data

    def _write(self, addr: int, val: int, size: int) -> None:
        self._attempt_write(addr, val.to_bytes(size, "little"))

    def write1(self, addr: int, val: int) -> None:
        self._write(addr, val, 1)

    def write2(self, addr: int, val: int) -> None:
        self._write(addr, val, 2)

    def write4(self, addr: int, val: int) -> None:
        self._write(addr, val, 4)

    def _read(self, addr: int, size: int) -> int:
        return int.from_bytes(self._attempt_read(addr, size), "little")

    def read1(self, addr: int) -> int:
        return self._read(addr, 1)

    def read2(self, addr: int) -> int:
        return self._read(addr, 2)

    def read4(self, addr: int) -> int:
        return self._read(addr, 4)


class CurverunnerCommSerial(CurverunnerComm):
    def __init__(self, ser: str | serial.Serial | None = None):
        """_summary_

        Args:
            ser (str | serial.Serial | None, optional): Serial connection specification. Can be:
                - str: Serial port path/name (e.g., '/dev/ttyUSB0' or 'COM3')
                - serial.Serial: Existing Serial object to use
                - None: Auto-discover first available CurveRunner device

        Raises:
            RuntimeError: If no CurveRunner devices are found during auto-discovery
        """

        if isinstance(ser, str):
            self._ser = serial.Serial(ser, 115200)
        elif ser is None:
            # Find first valid CurveRunner device
            devices = discover_devices_serial()
            if len(devices) == 0:
                raise RuntimeError("No CurveRunner devices found")

            self._ser = serial.Serial(devices[0], 115200)
        else:
            self._ser = ser

        self._lock = threading.RLock()

    def _error_recovery(self):
        self._ser.write(b"\n")
        self._ser.write(b"\n")
        self._ser.flush()
        time.sleep(0.1)
        self._ser.readall()

    def _exec_command(self, command: str) -> str:
        with self._lock:
            self._ser.write(command.encode() + b"\n")
            res = self._ser.readline().decode()
            if res.startswith("!") or res.startswith("?"):
                self._error_recovery()
                raise RuntimeError("Invalid command")
            else:
                return res

    def write1(self, addr: int, val: int) -> None:
        try:
            res = self._exec_command(f"W1,{addr},{val}")
            if not res.startswith("S"):
                raise RuntimeError("Failed to write")
        except:
            self._error_recovery()
            raise

    def write2(self, addr: int, val: int) -> None:
        try:
            res = self._exec_command(f"W2,{addr},{val}")
            if not res.startswith("S"):
                raise RuntimeError("Failed to write")
        except:
            self._error_recovery()
            raise

    def write4(self, addr: int, val: int) -> None:
        try:
            res = self._exec_command(f"W4,{addr},{val}")
            if not res.startswith("S"):
                raise RuntimeError("Failed to write")
        except:
            self._error_recovery()
            raise

    def read1(self, addr: int) -> int:
        try:
            res = self._exec_command(f"R1,{addr},0")
            return int(res)
        except:
            self._error_recovery()
            raise

    def read2(self, addr: int) -> int:
        try:
            res = self._exec_command(f"R2,{addr},0")
            return int(res)
        except:
            self._error_recovery()
            raise

    def read4(self, addr: int) -> int:
        try:
            res = self._exec_command(f"R4,{addr},0")
            return int(res)
        except:
            self._error_recovery()
            raise


class CurverunnerCommMock(CurverunnerComm):
    def __init__(self):
        self._regs = bytearray(256)

    def write1(self, addr: int, val: int) -> None:
        if addr < 0 or addr >= len(self._regs):
            return
        self._regs[addr] = val & 0xFF

    def write2(self, addr: int, val: int) -> None:
        if addr < 0 or addr + 1 >= len(self._regs):
            return

        self._regs[addr] = val & 0xFF
        self._regs[addr + 1] = (val >> 8) & 0xFF

    def write4(self, addr: int, val: int) -> None:
        if addr < 0 or addr + 3 >= len(self._regs):
            return

        self._regs[addr] = val & 0xFF
        self._regs[addr + 1] = (val >> 8) & 0xFF
        self._regs[addr + 2] = (val >> 16) & 0xFF
        self._regs[addr + 3] = (val >> 24) & 0xFF

    def read1(self, addr: int) -> int:
        return self._regs[addr]

    def read2(self, addr: int) -> int:
        return (self._regs[addr + 1] << 8) | self._regs[addr]

    def read4(self, addr: int) -> int:
        return (
            (self._regs[addr + 3] << 24)
            | (self._regs[addr + 2] << 16)
            | (self._regs[addr + 1] << 8)
            | self._regs[addr]
        )


def discover_devices_serial_by_id() -> List[Tuple[int, str]]:
    port_infos = comports()
    cr_list = []
    for port_info in port_infos:
        if port_info.serial_number is None:
            continue
        try:
            with serial.Serial(port_info.device, 115200, timeout=0.05, write_timeout=0.05) as ser:
                comm = CurverunnerCommSerial(ser)

                version = comm.read1(1)
                cr_list.append((version, port_info.device))
        except:
            continue

    return cr_list


def discover_devices_serial() -> List[str]:
    return [dev for _, dev in discover_devices_serial_by_id()]



INT16_MAX = (1 << 15) - 1

def signed_to_unsigned2(signed: int) -> int:
    return signed & 0xFFFF

def unsigned_to_signed2(unsigned: int) -> int:
    return (unsigned & ((1 << 15) - 1)) - (unsigned & (1 << 15))

def signed_to_unsigned4(signed: int) -> int:
    return signed & 0xFFFFFFFF

def unsigned_to_signed4(unsigned: int) -> int:
    return (unsigned & ((1 << 31) - 1)) - (unsigned & (1 << 31))

# arduino style map function
def map_value(val, from_low: float, from_high: float, to_low: float, to_high: float) -> float:
    return (val - from_low) / (from_high - from_low) * (to_high - to_low) + to_low


from .version import VERSION, VERSION_SHORT

from .comm import (
    CurverunnerComm,
    CurverunnerCommSerial,
    CurverunnerCommMock,
    discover_devices_serial,
    discover_devices_serial_by_id,
)
from .curverunner import Curverunner, CurverunnerMotor, CurverunnerServo

__version__ = VERSION
__version_short__ = VERSION_SHORT

__all__ = [
    "Curverunner",
    "CurverunnerMotor",
    "CurverunnerServo",
    "CurverunnerComm",
    "CurverunnerCommSerial",
    "CurverunnerCommMock",
    "discover_devices_serial",
    "discover_devices_serial_by_id",
]



"""Example script to control a CurveRunner device via USB serial communication.
"""

from curverunner import Curverunner, CurverunnerCommSerial
import time


# Create a CurveRunner instance using USB serial communication
# When no argument is given, the first available CurveRunner device is used
# I2C is recommended for usage with Raspberry Pi
comm = CurverunnerCommSerial()
# Create a CurveRunner instance using the communication interface
cr = Curverunner(comm)
print(f"Connected to CurveRunner with ID: {cr.get_device_id()}")

print("Setting servo 1 to 0 degrees")
cr.servo1.set_angle(0)
time.sleep(2)

print("Setting servo 1 to 90 degrees")
cr.servo1.set_angle(90)
time.sleep(2)

print("Setting servo 1 to 0 degrees")
cr.servo1.set_angle(0)



import struct
import time
from typing import List, Dict, Tuple

import serial
from serial.tools.list_ports import comports
import threading
import smbus2

from abc import ABC, abstractmethod

from curverunner.comm import CurverunnerComm, CurverunnerCommSerial, discover_devices_serial
from curverunner.util import INT16_MAX, map_value, unsigned_to_signed2, unsigned_to_signed4


class CurverunnerServo:
    SERVO_ANGLE_ADDRESSES = [3, 5, 7]

    SERVO_MIN_ANGLE = 0
    SERVO_MAX_ANGLE = 180

    SERVO_PORT_MIN_VAL = 0
    SERVO_PORT_MAX_VAL = 1800

    def __init__(self, comm: CurverunnerComm, servo_port: int = 1):
        self._comm = comm
        if servo_port < 1 or servo_port > 3:
            raise ValueError("Servo port must be between 1 and 3")
        self.servo_port = servo_port
        self._servo_addr = self.SERVO_ANGLE_ADDRESSES[servo_port - 1]

    def set_angle(self, degrees: float):
        """Move servo to the specified angle in degrees. \n

        Note that this method returns immediately and will not wait for the servo to reach the position.

        Args:
            degrees (float): The desired servo angle in degrees. Must be between 0 and 180 degrees.

        Raises:
            ValueError: If the degrees value is outside the valid servo angle range.
        """

        if degrees < self.SERVO_MIN_ANGLE or degrees > self.SERVO_MAX_ANGLE:
            raise ValueError(
                f"Servo angle must be between {self.SERVO_MIN_ANGLE} and {self.SERVO_MAX_ANGLE}"
            )
        val = int(
            map_value(
                degrees,
                self.SERVO_MIN_ANGLE,
                self.SERVO_MAX_ANGLE,
                self.SERVO_PORT_MIN_VAL,
                self.SERVO_PORT_MAX_VAL,
            )
        )
        self._comm.write2(self._servo_addr, val)

    def get_angle(self) -> float:
        """Gets the current servo angle in degrees.

        Returns:
            float: The current servo angle in degrees.
        """
        val = self._comm.read1(self._servo_addr)
        degrees = map_value(
            val,
            self.SERVO_PORT_MIN_VAL,
            self.SERVO_PORT_MAX_VAL,
            self.SERVO_MIN_ANGLE,
            self.SERVO_MAX_ANGLE,
        )
        return degrees


class CurverunnerMotor:
    REG_OFFSET_MOTOR_TYPE = 0
    REG_OFFSET_PULSES_PER_REV = 1
    REG_OFFSET_MOTOR_INVERTED = 3
    REG_OFFSET_ENCODER_INVERTED = 4
    REG_OFFSET_SLEW_RATE = 5
    REG_OFFSET_P = 7
    REG_OFFSET_I = 9
    REG_OFFSET_D = 11
    REG_OFFSET_FF = 13
    REG_OFFSET_TARGET = 15
    REG_OFFSET_CONTROL_MODE = 19
    REG_OFFSET_VEL = 20
    REG_OFFSET_POS = 22
    REG_OFFSET_PERCENT_OUTPUT = 26

    PIDF_SCALE = 1000

    MOTOR_TYPE_DISABLED = 0
    MOTOR_TYPE_OPEN_LOOP_BDC = 1
    MOTOR_TYPE_CLOSED_LOOP_BDC = 2
    MOTOR_TYPE_FIT0441 = 3

    CONTROL_MODE_DISABLED = 0
    CONTROL_MODE_PERCENT_OUTPUT = 1
    CONTROL_MODE_VELOCITY = 2
    CONTROL_MODE_POSITION = 3

    MOTOR_BASE_ADDRS = [15, 0]

    def __init__(self, comm: CurverunnerComm, motor_port: int = 1):
        if motor_port < 1 or motor_port > 2:
            raise ValueError("Motor port must be between 1 and 2")
        self.motor_port = motor_port
        self._base_addr = self.MOTOR_BASE_ADDRS[motor_port - 1]
        self._comm = comm

    def set_percent_out(self, percent_out: float):
        INT_16_MAX = (1 << 15) - 1
        if percent_out < -1:
            percent_out = -1
        elif percent_out > 1:
            percent_out = 1

        # convert to
        percent_out_write = int(percent_out * INT_16_MAX + INT_16_MAX)
        self._comm.write2(self._base_addr + self.REG_OFFSET_TARGET, percent_out_write)
        self._comm.write2(
            self._base_addr + self.REG_OFFSET_CONTROL_MODE, int(self.CONTROL_MODE_PERCENT_OUTPUT)
        )

    def set_target_velocity(self, velocity: int):
        vel = int(velocity)
        self._comm.write2(self._base_addr + self.REG_OFFSET_TARGET, vel)
        self._comm.write2(
            self._base_addr + self.REG_OFFSET_CONTROL_MODE, int(self.CONTROL_MODE_VELOCITY)
        )

    def get_position_raw(self) -> int:
        return unsigned_to_signed4(self._comm.read4(self._base_addr + self.REG_OFFSET_POS))

    def get_velocity_raw(self) -> int:
        return unsigned_to_signed2(self._comm.read2(self._base_addr + self.REG_OFFSET_VEL))

    def get_percent_out(self) -> float:
        return (
            unsigned_to_signed2(self._comm.read2(self._base_addr + self.REG_OFFSET_PERCENT_OUTPUT))
            / INT16_MAX
        )

    def set_p(self, p: float):
        self._comm.write2(self._base_addr + self.REG_OFFSET_P, int(p * self.PIDF_SCALE))

    def get_p(self) -> float:
        return self._comm.read2(self._base_addr + self.REG_OFFSET_P) / self.PIDF_SCALE

    def set_i(self, i: float):
        self._comm.write2(self._base_addr + self.REG_OFFSET_I, int(i * self.PIDF_SCALE))

    def get_i(self) -> float:
        return self._comm.read2(self._base_addr + self.REG_OFFSET_I) / self.PIDF_SCALE

    def set_d(self, d: float):
        self._comm.write2(self._base_addr + self.REG_OFFSET_D, int(d * self.PIDF_SCALE))

    def get_d(self) -> float:
        return self._comm.read2(self._base_addr + self.REG_OFFSET_D) / self.PIDF_SCALE

    def set_ff(self, f: float):
        self._comm.write2(self._base_addr + self.REG_OFFSET_FF, int(f * self.PIDF_SCALE))

    def get_ff(self) -> float:
        return self._comm.read2(self._base_addr + self.REG_OFFSET_FF) / self.PIDF_SCALE

    def get_motor_type(self) -> int:
        return self._comm.read1(self._base_addr + self.REG_OFFSET_MOTOR_TYPE)

    def set_motor_type(self, motor_type: int):
        if motor_type < 0 or motor_type > 3:
            raise RuntimeError("Motor type must be between 0 and 3")
        self._comm.write1(self._base_addr + self.REG_OFFSET_MOTOR_TYPE, int(motor_type))

    def set_encoder_inverted(self, inverted: bool):
        self._comm.write1(self._base_addr + self.REG_OFFSET_ENCODER_INVERTED, int(inverted))

    def get_encoder_inverted(self) -> bool:
        return self._comm.read1(self._base_addr + self.REG_OFFSET_ENCODER_INVERTED) == 1

    def set_motor_inverted(self, inverted: bool):
        self._comm.write1(self._base_addr + self.REG_OFFSET_MOTOR_INVERTED, int(inverted))

    def get_motor_inverted(self) -> bool:
        return self._comm.read1(self._base_addr + self.REG_OFFSET_MOTOR_INVERTED) == 1


class Curverunner:
    REG_VERSION = 0
    REG_DEVICE_ID = 1

    REG_PARAM1 = 71
    REG_PARAM2 = 73
    REG_COMMAND = 75

    COMMAND_SAVE_CONFIG = 1
    COMMAND_RELOAD_CONFIG = 2
    COMMAND_FACTORY_RESET = 3
    COMMAND_SET_DEVICE_ID = 4

    def __init__(self, comm: CurverunnerComm):
        self.comm = comm
        self.servo1 = CurverunnerServo(comm, 1)
        self.servo2 = CurverunnerServo(comm, 2)
        self.servo3 = CurverunnerServo(comm, 3)

        self.motor1 = CurverunnerMotor(comm, 1)

    def get_device_id(self) -> int:
        return self.comm.read1(self.REG_DEVICE_ID)

    def set_device_id(self, device_id: int):
        self.comm.write1(self.REG_DEVICE_ID, device_id)
        self.comm.write1(self.REG_COMMAND, self.COMMAND_SET_DEVICE_ID)

    def save_config(self):
        self.comm.write1(self.REG_COMMAND, self.COMMAND_SAVE_CONFIG)

    def reload_config(self):
        self.comm.write1(self.REG_COMMAND, self.COMMAND_RELOAD_CONFIG)

    def factory_reset(self):
        self.comm.write1(self.REG_COMMAND, self.COMMAND_FACTORY_RESET)
